/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.objectweb.petals.tools.rmi.server.remote.implementations;

import java.net.URI;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;

import javax.jbi.messaging.InOnly;
import javax.jbi.messaging.InOptionalOut;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.RobustInOnly;
import javax.xml.namespace.QName;

import org.objectweb.petals.tools.rmi.remote.RemoteMessageExchangeFactory;
import org.objectweb.petals.tools.rmi.server.remote.implementations.RemoteComponentContextClientFactoryImpl.SharedComponentData;
import org.objectweb.petals.tools.rmi.server.util.Convert;
import org.ow2.petals.jbi.servicedesc.endpoint.Location;

/**
 * the RMI Message Exchange Factory.
 * 
 * @author Nicolas Salatge - EBM WebSourcing
 */
public class RemoteMessageExchangeFactoryImpl extends UnicastRemoteObject implements RemoteMessageExchangeFactory {

    /**
     * the serialVersionUID.
     */
    private static final long serialVersionUID = -7802199904640493500L;

    /**
     * the concrete message exchange factory.
     */
    private final MessageExchangeFactory concreteMef;

    /**
     * The RMI client id in the JBI environment to associate the message exchanges to the right RMI client of the SE
     * RMI.
     */
    private final String rmiClientId;

    private final SharedComponentData sharedData;

    private final Location location;

    public RemoteMessageExchangeFactoryImpl(final SharedComponentData sharedData,
            final MessageExchangeFactory concreteMef, final String rmiClientId) throws RemoteException {
        this.sharedData = sharedData;
        this.concreteMef = concreteMef;
        this.rmiClientId = rmiClientId;
        this.location = new Location(sharedData.componentContext.getComponentName(), rmiClientId);
    }

    @Override
    public MessageExchange createExchange(final QName serviceName, final QName operationName) throws MessagingException {
        try {
            // Create the concrete message exchange
            final MessageExchange meC = this.concreteMef.createExchange(serviceName, operationName);
            meC.setProperty(Convert.RMI_CLIENT_ID, this.rmiClientId);

            // Convert into fake message exchange
            final MessageExchange meF = Convert.createFakeMessageExchangeFromConcreteMessageExchange(meC,
                    this.location);

            // Store the messages
            this.sharedData.messages.put(meF.getExchangeId(), meC);

            return meF;
        } catch (final MessagingException e) {
            throw Convert.convertMessagingException(e);
        }
    }

    @Override
    public MessageExchange createExchange(final URI pattern) throws MessagingException {
        try {
            // Create the concrete message exchange
            final MessageExchange meC = this.concreteMef.createExchange(pattern);
            meC.setProperty(Convert.RMI_CLIENT_ID, this.rmiClientId);

            // Convert into fake message exchange
            final MessageExchange meF = Convert.createFakeMessageExchangeFromConcreteMessageExchange(meC,
                    this.location);

            // Store the messages
            this.sharedData.messages.put(meF.getExchangeId(), meC);

            return meF;
        } catch (final MessagingException e) {
            throw Convert.convertMessagingException(e);
        }
    }

    @Override
    public InOnly createInOnlyExchange() throws MessagingException {
        try {
            // Create the concrete message exchange
            final InOnly meC = this.concreteMef.createInOnlyExchange();

            meC.setProperty(Convert.RMI_CLIENT_ID, this.rmiClientId);

            // Convert into fake message exchange
            final InOnly meF = Convert.createFakeMessageExchangeFromConcreteMessageExchange(meC, this.location);

            // Store the messages
            this.sharedData.messages.put(meF.getExchangeId(), meC);

            return meF;
        } catch (final MessagingException e) {
            throw Convert.convertMessagingException(e);
        }
    }

    @Override
    public InOptionalOut createInOptionalOutExchange() throws MessagingException {
        try {
            // Create the concrete message exchange
            final InOptionalOut meC = this.concreteMef.createInOptionalOutExchange();

            meC.setProperty(Convert.RMI_CLIENT_ID, this.rmiClientId);

            // Convert into fake message exchange
            final InOptionalOut meF = Convert.createFakeMessageExchangeFromConcreteMessageExchange(meC, this.location);

            // Store the messages
            this.sharedData.messages.put(meF.getExchangeId(), meC);

            return meF;
        } catch (final MessagingException e) {
            throw Convert.convertMessagingException(e);
        }
    }

    @Override
    public InOut createInOutExchange() throws MessagingException {
        try {
            // Create the concrete message exchange
            final InOut meC = this.concreteMef.createInOutExchange();
            meC.setProperty(Convert.RMI_CLIENT_ID, this.rmiClientId);

            // Convert into fake message exchange
            final InOut meF = Convert.createFakeMessageExchangeFromConcreteMessageExchange(meC, this.location);

            // Store the messages
            this.sharedData.messages.put(meF.getExchangeId(), meC);

            return meF;
        } catch (final MessagingException e) {
            throw Convert.convertMessagingException(e);
        }
    }

    @Override
    public RobustInOnly createRobustInOnlyExchange() throws MessagingException {
        try {
            // Create the concrete message exchange
            final RobustInOnly meC = this.concreteMef.createRobustInOnlyExchange();
            meC.setProperty(Convert.RMI_CLIENT_ID, this.rmiClientId);

            // Convert into fake message exchange
            final RobustInOnly meF = Convert.createFakeMessageExchangeFromConcreteMessageExchange(meC, this.location);

            // Store the messages
            this.sharedData.messages.put(meF.getExchangeId(), meC);

            return meF;
        } catch (final MessagingException e) {
            throw Convert.convertMessagingException(e);
        }
    }
}
