/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.objectweb.petals.tools.rmi.server.util;

import java.util.HashSet;
import java.util.Set;

import javax.jbi.JBIException;
import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.InOnly;
import javax.jbi.messaging.InOptionalOut;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.messaging.RobustInOnly;
import javax.jbi.servicedesc.ServiceEndpoint;

import org.ow2.easywsdl.wsdl.api.abstractItf.AbsItfOperation.MEPPatternConstants;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeWrapper;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;
import org.ow2.petals.jbi.messaging.exchange.impl.MessageExchangeImpl;
import org.ow2.petals.jbi.messaging.exchange.types.InOnlyImpl;
import org.ow2.petals.jbi.messaging.exchange.types.InOptionalOutImpl;
import org.ow2.petals.jbi.messaging.exchange.types.InOutImpl;
import org.ow2.petals.jbi.messaging.exchange.types.RobustInOnlyImpl;
import org.ow2.petals.jbi.servicedesc.endpoint.Location;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint.EndpointType;
import org.ow2.petals.jbi.servicedesc.endpoint.impl.JBIServiceEndpointImpl;

import com.ebmwebsourcing.easycommons.lang.UncheckedException;

/**
 * Static class that allow to convert fake client rmi object to concrete petals object and vice and versa.
 * 
 * @author Nicolas Salatge - EBM WebSourcing
 */
public final class Convert {

    public static final String RMI_CLIENT_ID = "petals-rmi-server.rmi-client.identifier";

    private Convert() {

    }

    public static ServiceEndpoint convertConcreteServiceEndpointToFakeServiceEndpoint(final ServiceEndpoint concrete) {
        return new JBIServiceEndpointImpl(EndpointType.INTERNAL, concrete.getEndpointName(), concrete.getServiceName(),
                concrete.getInterfaces().clone(), new Location("fake-RMI-cont", "fake-RMI-comp"));
    }

    /**
     * Convert a concrete message exchange to a fake message exchange.
     */
    public static <T extends MessageExchange> T createFakeMessageExchangeFromConcreteMessageExchange(final T concrete,
            final Location location)
            throws MessagingException {

        // Create Fake Message Exchange
        final T fake = Convert.createFakeMessage(concrete, location);
        
        // let's disable any checks (because it could be either from a clean one or an accepted one from a different
        // role and it's easier that to try to mirror the roles and mep order of actions)
        final MessageExchangeWrapper wrapper = (MessageExchangeWrapper) fake;
        final PetalsMessageExchange exchange = wrapper.removeOwnership();
        exchange.setDoChecks(false);

        @SuppressWarnings("unchecked")
        final Set<String> props = concrete.getPropertyNames();
        for (final String property : props) {
            exchange.setProperty(property, concrete.getProperty(property));
        }

        // Set all values in concrete message exchange to fake message exchange
        if (concrete.getEndpoint() != null) {
            exchange.setEndpoint(convertConcreteServiceEndpointToFakeServiceEndpoint(concrete.getEndpoint()));
        }

        if (concrete.getError() != null) {
            exchange.setError(Convert.convertException(concrete.getError()));
        }

        if (concrete.getFault() != null) {
            exchange.setFault(Convert.convertConcreteToFakeNormalizedMessage(exchange, concrete.getFault()));
        }

        if (concrete.getMessage(PetalsMessageExchange.IN_MSG) != null) {
            exchange.setMessage(Convert.convertConcreteToFakeNormalizedMessage(exchange,
                    concrete.getMessage(PetalsMessageExchange.IN_MSG)), PetalsMessageExchange.IN_MSG);
        }

        if (concrete.getMessage(PetalsMessageExchange.OUT_MSG) != null) {
            exchange.setMessage(Convert.convertConcreteToFakeNormalizedMessage(exchange,
                    concrete.getMessage(PetalsMessageExchange.OUT_MSG)), PetalsMessageExchange.OUT_MSG);
        }

        exchange.setInterfaceName(concrete.getInterfaceName());
        exchange.setOperation(concrete.getOperation());
        exchange.setService(concrete.getService());
        exchange.setStatus(concrete.getStatus());

        exchange.setDoChecks(false);
        wrapper.returnOwnership(exchange);

        return fake;
    }

    /**
     * Convert a concrete messaging exception to fake messaging exception.
     */
    public static final JBIException convertJBIException(final JBIException e) {
        final JBIException me = new JBIException(e.getMessage());
        me.setStackTrace(e.getStackTrace());
        return me;
    }

    /**
     * Convert a concrete messaging exception to fake messaging exception.
     */
    public static final MessagingException convertMessagingException(final MessagingException e) {
        final MessagingException me = new MessagingException(e.getMessage());
        me.setStackTrace(e.getStackTrace());
        return me;
    }

    /**
     * Convert an error exception to a RMI exception.
     */
    public static Exception convertException(final Exception e) {
        final Exception exception = new Exception(e.getMessage());
        exception.setStackTrace(e.getStackTrace());
        return exception;
    }

    /**
     * Convert an error exception to a RMI exception.
     */
    public static RuntimeException convertRuntimeException(final Exception e) {
        final RuntimeException exception = new RuntimeException(e.getMessage());
        exception.setStackTrace(e.getStackTrace());
        return exception;
    }

    /**
     * Set Concrete Value of Message Exchange in fake message exchange.
     */
    public static void setConcreteValueMEToFakeValueME(final MessageExchange concrete, final MessageExchange fake)
            throws MessagingException {

        // let's disable the checks because we are copying from an exchange sent by another role
        final MessageExchangeWrapper wrapper = (MessageExchangeWrapper) fake;
        final PetalsMessageExchange exchange = wrapper.removeOwnership();
        exchange.setDoChecks(false);

        updateProperties(concrete, exchange);

        if (concrete.getStatus() == ExchangeStatus.DONE) {
            // Status done
            exchange.setStatus(ExchangeStatus.DONE);
        } else if (concrete.getStatus() == ExchangeStatus.ERROR) {
            // Status error (we convert it to avoid classpath problems!)
            exchange.setError(convertException(concrete.getError()));
            // theoretically, we don't need to do that, but sometimes the message is in error without any error set!
            exchange.setStatus(ExchangeStatus.ERROR);
        } else {
            // Status active

            // Set all values in fake message exchange to concrete message exchange
            exchange.setStatus(ExchangeStatus.ACTIVE);

            exchange.setOperation(concrete.getOperation());

            if (concrete.getEndpoint() != null && exchange.getEndpoint() == null) {
                exchange.setEndpoint(convertConcreteServiceEndpointToFakeServiceEndpoint(concrete.getEndpoint()));
            }

            if (exchange.getInterfaceName() == null) {
                exchange.setInterfaceName(concrete.getInterfaceName());
            }

            if (exchange.getService() == null) {
                exchange.setService(concrete.getService());
            }

            if (concrete.getFault() != null && exchange.getFault() == null) {
                exchange.setFault(Convert.convertConcreteToFakeNormalizedMessage(exchange, concrete.getFault()));
            }

            if (concrete.getMessage(PetalsMessageExchange.IN_MSG) != null
                    && exchange.getMessage(PetalsMessageExchange.IN_MSG) == null) {
                exchange.setMessage(Convert.convertConcreteToFakeNormalizedMessage(exchange,
                        concrete.getMessage(PetalsMessageExchange.IN_MSG)), PetalsMessageExchange.IN_MSG);
            }

            if (concrete.getMessage(PetalsMessageExchange.OUT_MSG) != null
                    && exchange.getMessage(PetalsMessageExchange.OUT_MSG) == null) {
                exchange.setMessage(Convert.convertConcreteToFakeNormalizedMessage(exchange,
                        concrete.getMessage(PetalsMessageExchange.OUT_MSG)), PetalsMessageExchange.OUT_MSG);
            }
        }

        exchange.setDoChecks(true);
        wrapper.returnOwnership(exchange);
    }

    /**
     * Set Fake Value of Message Exchange in concrete message exchange.
     */
    public static void setFakeValueMEToConcreteValueME(final ComponentContext cc, final MessageExchange fake,
            final MessageExchange concrete) throws MessagingException {

        updateProperties(fake, concrete);

        if (fake.getStatus() == ExchangeStatus.DONE) {
            // Status done
            concrete.setStatus(ExchangeStatus.DONE);
        } else if (fake.getStatus() == ExchangeStatus.ERROR) {
            // Status error
            concrete.setError(fake.getError());
            // theoretically, we don't need to do that, but sometimes the message is in error without any error set!
            concrete.setStatus(ExchangeStatus.ERROR);
        } else {
            // Status active
            concrete.setStatus(ExchangeStatus.ACTIVE);

            concrete.setOperation(fake.getOperation());

            // Set all values in fake message exchange to concrete message exchange
            if (fake.getEndpoint() != null && concrete.getEndpoint() == null) {
                concrete.setEndpoint(
                        cc.getEndpoint(fake.getEndpoint().getServiceName(), fake.getEndpoint().getEndpointName()));
            }

            if (concrete.getInterfaceName() == null) {
                concrete.setInterfaceName(fake.getInterfaceName());
            }

            if (concrete.getService() == null) {
                concrete.setService(fake.getService());
            }

            if (fake.getFault() != null && concrete.getFault() == null) {
                concrete.setFault(Convert.convertFakeToConcreteNormalizedMessage(concrete, fake.getFault()));
            }

            if (fake.getMessage(PetalsMessageExchange.IN_MSG) != null
                    && concrete.getMessage(PetalsMessageExchange.IN_MSG) == null) {
                concrete.setMessage(Convert.convertFakeToConcreteNormalizedMessage(concrete,
                        fake.getMessage(PetalsMessageExchange.IN_MSG)), PetalsMessageExchange.IN_MSG);
            }

            if (fake.getMessage(PetalsMessageExchange.OUT_MSG) != null
                    && concrete.getMessage(PetalsMessageExchange.OUT_MSG) == null) {
                concrete.setMessage(Convert.convertFakeToConcreteNormalizedMessage(concrete,
                        fake.getMessage(PetalsMessageExchange.OUT_MSG)), PetalsMessageExchange.OUT_MSG);
            }
        }
    }

    private static void updateProperties(final MessageExchange from, final MessageExchange to) {
        // need a copy to prevent concurrent modifications
        @SuppressWarnings("unchecked")
        final Set<String> oldProps = new HashSet<>(to.getPropertyNames());
        // let's remove the old properties
        for (final String property : oldProps) {
            to.setProperty(property, null);
        }

        @SuppressWarnings("unchecked")
        final Set<String> props = from.getPropertyNames();

        // and replace with the ones from the fake
        for (final String property : props) {
            to.setProperty(property, from.getProperty(property));
        }
    }

    /**
     * convert a concrete NormalizedMessage to a fake NormalizedMessage.
     */
    private static final <T extends NormalizedMessage> T convertConcreteToFakeNormalizedMessage(
            final MessageExchange fakeME, final T concrete) throws MessagingException {
        NormalizedMessage fake = null;
        if (concrete != null) {
            if (concrete instanceof Fault) {
                // Create a fake FaultMessage
                fake = fakeME.createFault();
            } else {
                // Create a fake NormalizedMessage
                fake = fakeME.createMessage();
            }

            // Set all values in fake normalized message to concrete normalized
            // message
            fake.setContent(concrete.getContent());

            for (final Object property : concrete.getPropertyNames()) {
                fake.setProperty((String) property, concrete.getProperty((String) property));
            }

            fake.setSecuritySubject(concrete.getSecuritySubject());

            for (final Object attachmentName : concrete.getAttachmentNames()) {
                fake.addAttachment((String) attachmentName, concrete.getAttachment((String) attachmentName));
            }
        }

        @SuppressWarnings("unchecked")
        final T res = (T) fake;

        return res;
    }

    /**
     * convert a fake NormalizedMessage to a concrete NormalizedMessage.
     */
    private static final <T extends NormalizedMessage> T convertFakeToConcreteNormalizedMessage(
            final MessageExchange concreteME, final T fake) throws MessagingException {
        NormalizedMessage concrete = null;

        if (fake != null) {
            if (fake instanceof Fault) {
                // Create a concrete FaultMessage
                concrete = concreteME.createFault();
            } else {
                // Create a concrete NormalizedMessage
                concrete = concreteME.createMessage();
            }

            // Set all values in fake normalized message to concrete normalized
            // message
            concrete.setContent(fake.getContent());

            if (fake.getPropertyNames() != null) {
                for (final Object property : fake.getPropertyNames()) {
                    concrete.setProperty((String) property, fake.getProperty((String) property));
                }
            }

            concrete.setSecuritySubject(fake.getSecuritySubject());

            if (fake.getAttachmentNames() != null) {
                for (final Object attachmentName : fake.getAttachmentNames()) {
                    concrete.addAttachment((String) attachmentName, fake.getAttachment((String) attachmentName));
                }
            }
        }

        @SuppressWarnings("unchecked")
        final T res = (T) concrete;
        return res;
    }

    /**
     * Create an empty fake message.
     */
    private static final <T extends MessageExchange> T createFakeMessage(final T concrete, final Location location) {

        final PetalsMessageExchange me = new MessageExchangeImpl(concrete.getExchangeId(), concrete.getPattern(),
                location);

        final MessageExchange fake;

        // careful, RobustInOnly must be tested before InOnly, and InOptionalOut before InOut, because of inheritance of
        // the implementations!
        if (concrete instanceof RobustInOnly) {
            fake = new RobustInOnlyImpl(me);
            if (!MEPPatternConstants.ROBUST_IN_ONLY.equals(concrete.getPattern())) {
                throw new UncheckedException("There must be a bug somewhere, received a RobustInOnly exchange with mep "
                        + concrete.getPattern());
            }
        } else if (concrete instanceof InOnly) {
            fake = new InOnlyImpl(me);
            if (!MEPPatternConstants.IN_ONLY.equals(concrete.getPattern())) {
                throw new UncheckedException(
                        "There must be a bug somewhere, received a InOnly exchange with mep " + concrete.getPattern());
            }
        } else if (concrete instanceof InOptionalOut) {
            fake = new InOptionalOutImpl(me);
            if (!MEPPatternConstants.IN_OPTIONAL_OUT.equals(concrete.getPattern())) {
                throw new UncheckedException(
                        "There must be a bug somewhere, received a InOptionalOut exchange with mep "
                                + concrete.getPattern());
            }
        } else if (concrete instanceof InOut) {
            fake = new InOutImpl(me);
            if (!MEPPatternConstants.IN_OUT.equals(concrete.getPattern())) {
                throw new UncheckedException(
                        "There must be a bug somewhere, received a InOut exchange with mep " + concrete.getPattern());
            }
        } else {
            fake = new MessageExchangeWrapper(me);
        }

        @SuppressWarnings("unchecked")
        final T res = (T) fake;

        return res;
    }
}
